@@ -1,19 +1,15 @@
-use log::{debug, log_enabled};
-use std::f64;
+use alloc::vec::Vec;
 
-use crate::deflate::calculate_block_size_auto_type;
-use crate::lz77::{Lz77Store, ZopfliBlockState};
-use crate::Options;
+use log::{debug, log_enabled};
+
+use crate::{cache::NoCache, deflate::calculate_block_size_auto_type, lz77::Lz77Store};
 
 /// Finds minimum of function `f(i)` where `i` is of type `usize`, `f(i)` is of type
 /// `f64`, `i` is in range `start-end` (excluding `end`).
 /// Returns the index to the minimum and the minimum value.
-fn find_minimum<F>(f: F, start: usize, end: usize) -> (usize, f64)
-where
-    F: Fn(usize) -> f64,
-{
+fn find_minimum<F: Fn(usize) -> f64>(f: F, start: usize, end: usize) -> (usize, f64) {
     if end - start < 1024 {
-        let mut best = f64::MAX;
+        let mut best = f64::INFINITY;
         let mut result = start;
         for i in start..end {
             let v = f(i);
@@ -27,17 +23,17 @@ where
         /* Try to find minimum faster by recursively checking multiple points. */
         let mut start = start;
         let mut end = end;
-        let num = 9; /* Good value: 9. ?!?!?!?! */
-        let mut p = vec![0; num];
-        let mut vp = vec![0.0; num];
-        let mut lastbest = f64::MAX;
+        const NUM: usize = 9; /* Good value: 9. ?!?!?!?! */
+        let mut p = [0; NUM];
+        let mut vp = [0.0; NUM];
+        let mut lastbest = f64::INFINITY;
         let mut pos = start;
 
-        while end - start > num {
+        while end - start > NUM {
             let mut besti = 0;
-            let mut best = f64::MAX;
-            let multiplier = (end - start) / (num + 1);
-            for i in 0..num {
+            let mut best = f64::INFINITY;
+            let multiplier = (end - start) / (NUM + 1);
+            for i in 0..NUM {
                 p[i] = start + (i + 1) * multiplier;
                 vp[i] = f(p[i]);
                 if vp[i] < best {
@@ -50,7 +46,7 @@ where
             }
 
             start = if besti == 0 { start } else { p[besti - 1] };
-            end = if besti == num - 1 { end } else { p[besti + 1] };
+            end = if besti == NUM - 1 { end } else { p[besti + 1] };
 
             pos = p[besti];
             lastbest = best;
@@ -218,7 +214,6 @@ pub fn blocksplit_lz77(lz77: &Lz77Store, maxblocks: u16, splitpoints: &mut Vec<u
 /// npoints: pointer to amount of splitpoints, for the dynamic array. The amount of
 ///   blocks is the amount of splitpoitns + 1.
 pub fn blocksplit(
-    options: &Options,
     in_data: &[u8],
     instart: usize,
     inend: usize,
@@ -231,8 +226,7 @@ pub fn blocksplit(
     /* Unintuitively, Using a simple LZ77 method here instead of lz77_optimal
     results in better blocks. */
     {
-        let mut state = ZopfliBlockState::new_without_cache(options, instart, inend);
-        store.greedy(&mut state, in_data, instart, inend);
+        store.greedy(&mut NoCache, in_data, instart, inend);
     }
 
     let mut lz77splitpoints = Vec::with_capacity(maxblocks as usize);
