@@ -1,167 +1,167 @@
+use alloc::vec::Vec;
+use core::{cmp, iter};
+
 use log::{debug, log_enabled};
-use std::cmp;
-use std::io::{self, Read, Write};
 
-use crate::blocksplitter::{blocksplit, blocksplit_lz77};
-use crate::iter::ToFlagLastIterator;
-use crate::katajainen::length_limited_code_lengths;
-use crate::lz77::{LitLen, Lz77Store, ZopfliBlockState};
-use crate::squeeze::{lz77_optimal, lz77_optimal_fixed};
-use crate::symbols::{
-    get_dist_extra_bits, get_dist_extra_bits_value, get_dist_symbol, get_dist_symbol_extra_bits,
-    get_length_extra_bits, get_length_extra_bits_value, get_length_symbol,
-    get_length_symbol_extra_bits,
+use crate::{
+    blocksplitter::{blocksplit, blocksplit_lz77},
+    cache::ZopfliLongestMatchCache,
+    iter::ToFlagLastIterator,
+    katajainen::length_limited_code_lengths,
+    lz77::{LitLen, Lz77Store},
+    squeeze::{lz77_optimal, lz77_optimal_fixed},
+    symbols::{
+        get_dist_extra_bits, get_dist_extra_bits_value, get_dist_symbol,
+        get_dist_symbol_extra_bits, get_length_extra_bits, get_length_extra_bits_value,
+        get_length_symbol, get_length_symbol_extra_bits,
+    },
+    tree::lengths_to_symbols,
+    util::{ZOPFLI_NUM_D, ZOPFLI_NUM_LL, ZOPFLI_WINDOW_SIZE},
+    Error, Options, Write,
 };
-use crate::tree::lengths_to_symbols;
-use crate::util::{
-    read_to_fill, ZOPFLI_MASTER_BLOCK_SIZE, ZOPFLI_NUM_D, ZOPFLI_NUM_LL, ZOPFLI_WINDOW_SIZE,
-};
-use crate::Options;
 
-/// Compresses according to the deflate specification and append the compressed
-/// result to the output.
+/// A DEFLATE encoder powered by the Zopfli algorithm that compresses data written
+/// to it to the specified sink. Most users will find using [`compress`](crate::compress)
+/// easier and more performant.
 ///
-/// `options`: global program options
-/// `btype`: the deflate block type. Use Dynamic for best compression.
-///   - Uncompressed: non compressed blocks (00)
-///   - Fixed: blocks with fixed tree (01)
-///   - Dynamic: blocks with dynamic tree (10)
-/// `in_data`: the input bytes
-/// `out`: pointer to the dynamic output array to which the result is appended. Must
-///   be freed after use.
-pub fn deflate<R, W>(options: &Options, btype: BlockType, mut in_data: R, out: W) -> io::Result<()>
-where
-    R: Read,
-    W: Write,
-{
-    // The streaming compression algorithm implemented in this function is based on a sliding buffer
-    // that is used and structured as shown in the following ASCII art:
-    //
-    //             0 ┌────────┬───────┬─ ─ ─ ─ ─ ┐
-    // Iteration 0   │ Window │ Block │ Sentinel │
-    //               │  data  │ data  │   byte   │
-    //               └────────┴───────┴─ ─ ─ ─ ─ ┘ WINDOW_SIZE + MASTER_BLOCK_SIZE + 1
-    //                             \            │
-    //      Copy last WINDOW_SIZE   \           │
-    //           block bytes         \          │ Copy sentinel byte to beginning of block data
-    //                                V         V
-    //                               ┌────────┬───────┬─ ─ ─ ─ ─ ┐
-    // Iteration 1                   │ Window │ Block │ Sentinel │
-    //                               │  data  │ data  │   byte   │
-    //                               └────────┴───────┴─ ─ ─ ─ ─ ┘
-    //
-    // The input is read in blocks of MASTER_BLOCK_SIZE + 1 bytes. The last byte is used as
-    // a sentinel to know whether the stream ends at this master block without buffering
-    // previous master blocks. On the beginning of each iteration, the previous sentinel
-    // byte is copied to the beginning of the block data, to consume it for compression.
-    // If we read a full master block plus a new sentinel byte, there is a next master block
-    // at least one byte long, so the just read master block is non-last. If we can't read a
-    // full master block plus a new sentinel byte due to input EOF, the stream ends at this
-    // partial or full master block, so we know it is the last one.
-    //
-    // In addition to the above, a sliding window of the last WINDOW_SIZE bytes of the
-    // previous master block is kept, to allow the DEFLATE stream to reference data from
-    // the former master block, which is key to achieve the expected compression rates.
+/// The data will be compressed as soon as possible, without trying to fill a
+/// backreference window. As a consequence, frequent short writes may cause more
+/// DEFLATE blocks to be emitted with less optimal Huffman trees, which can hurt
+/// compression and runtime. If they are a concern, short writes can be conveniently
+/// dealt with by wrapping this encoder with a [`BufWriter`](std::io::BufWriter), as done
+/// by the [`new_buffered`](DeflateEncoder::new_buffered) method. An adequate write size
+/// would be >32 KiB, which allows the second complete chunk to leverage a full-sized
+/// backreference window.
+pub struct DeflateEncoder<W: Write> {
+    options: Options,
+    btype: BlockType,
+    have_chunk: bool,
+    chunk_start: usize,
+    window_and_chunk: Vec<u8>,
+    bitwise_writer: Option<BitwiseWriter<W>>,
+}
 
-    const ZOPFLI_WINDOW_AND_BLOCK_SIZE: usize = ZOPFLI_WINDOW_SIZE + ZOPFLI_MASTER_BLOCK_SIZE;
-
-    let mut buffer = vec![0; ZOPFLI_WINDOW_AND_BLOCK_SIZE + 1];
-
-    let mut have_window = false;
-    let mut sentinel_byte = None;
-
-    let mut bitwise_writer = BitwiseWriter::new(out);
-
-    loop {
-        // Inject previous sentinel byte as block data we've just read in this iteration.
-        let sentinel_byte_count = if let Some(sentinel_byte) = sentinel_byte.take() {
-            buffer[ZOPFLI_WINDOW_SIZE] = sentinel_byte;
-            1
-        } else {
-            0
-        };
-
-        // Read a new master block of data, plus a new sentinel byte.
-        match read_to_fill(
-            &mut in_data,
-            &mut buffer[ZOPFLI_WINDOW_SIZE + sentinel_byte_count..],
-        )? {
-            (true, _) => {
-                // We have read a full master block and a sentinel byte.
-
-                let (in_data, instart, inend) = if have_window {
-                    (
-                        &buffer[..ZOPFLI_WINDOW_AND_BLOCK_SIZE],
-                        ZOPFLI_WINDOW_SIZE,
-                        ZOPFLI_WINDOW_AND_BLOCK_SIZE,
-                    )
-                } else {
-                    (
-                        &buffer[ZOPFLI_WINDOW_SIZE..ZOPFLI_WINDOW_AND_BLOCK_SIZE],
-                        0,
-                        ZOPFLI_MASTER_BLOCK_SIZE,
-                    )
-                };
-
-                // The next block is at least one byte long due to the sentinel, so we know this block is not the last.
-                deflate_part(
-                    options,
-                    btype,
-                    false,
-                    in_data,
-                    instart,
-                    inend,
-                    &mut bitwise_writer,
-                )?;
-
-                // Update the DEFLATE sliding window with the latest bytes we've just read.
-                buffer.copy_within(
-                    ZOPFLI_WINDOW_AND_BLOCK_SIZE - ZOPFLI_WINDOW_SIZE..ZOPFLI_WINDOW_AND_BLOCK_SIZE,
-                    0,
-                );
-                have_window = true;
-
-                sentinel_byte = Some(buffer[buffer.len() - 1]);
-            }
-            (false, bytes_read) => {
-                // Due to EOF, we either have read a partial master block, or a full block without a sentinel byte.
-                // In both cases, no more blocks will follow.
-
-                let read_block_size = sentinel_byte_count + bytes_read;
-                let (in_data, instart, inend) = if have_window {
-                    (
-                        &buffer[..ZOPFLI_WINDOW_SIZE + read_block_size],
-                        ZOPFLI_WINDOW_SIZE,
-                        ZOPFLI_WINDOW_SIZE + read_block_size,
-                    )
-                } else {
-                    (
-                        &buffer[ZOPFLI_WINDOW_SIZE..ZOPFLI_WINDOW_SIZE + read_block_size],
-                        0,
-                        read_block_size,
-                    )
-                };
-
-                deflate_part(
-                    options,
-                    btype,
-                    true,
-                    in_data,
-                    instart,
-                    inend,
-                    &mut bitwise_writer,
-                )?;
-
-                break;
-            }
+impl<W: Write> DeflateEncoder<W> {
+    /// Creates a new Zopfli DEFLATE encoder that will operate according to the
+    /// specified options.
+    pub fn new(options: Options, btype: BlockType, sink: W) -> Self {
+        DeflateEncoder {
+            options,
+            btype,
+            have_chunk: false,
+            chunk_start: 0,
+            window_and_chunk: Vec::with_capacity(ZOPFLI_WINDOW_SIZE),
+            bitwise_writer: Some(BitwiseWriter::new(sink)),
         }
     }
 
-    bitwise_writer.finish_partial_bits()
+    /// Creates a new Zopfli DEFLATE encoder that operates according to the
+    /// specified options and is wrapped with a buffer to guarantee that
+    /// data is compressed in large chunks, which is necessary for decent
+    /// performance and good compression ratio.
+    #[cfg(feature = "std")]
+    pub fn new_buffered(options: Options, btype: BlockType, sink: W) -> std::io::BufWriter<Self> {
+        std::io::BufWriter::with_capacity(
+            crate::util::ZOPFLI_MASTER_BLOCK_SIZE,
+            Self::new(options, btype, sink),
+        )
+    }
+
+    /// Encodes any pending chunks of data and writes them to the sink,
+    /// consuming the encoder and returning the wrapped sink. The sink
+    /// will have received a complete DEFLATE stream when this method
+    /// returns.
+    ///
+    /// The encoder is automatically [`finish`](Self::finish)ed when
+    /// dropped, but explicitly finishing it with this method allows
+    /// handling I/O errors.
+    pub fn finish(mut self) -> Result<W, Error> {
+        self._finish().map(|sink| sink.unwrap())
+    }
+
+    /// Compresses the chunk stored at `window_and_chunk`. This includes
+    /// a rolling window of the last `ZOPFLI_WINDOW_SIZE` data bytes, if
+    /// available.
+    #[inline]
+    fn compress_chunk(&mut self, is_last: bool) -> Result<(), Error> {
+        deflate_part(
+            &self.options,
+            self.btype,
+            is_last,
+            &self.window_and_chunk,
+            self.chunk_start,
+            self.window_and_chunk.len(),
+            self.bitwise_writer.as_mut().unwrap(),
+        )
+    }
+
+    /// Sets the next chunk that will be compressed by the next
+    /// call to `compress_chunk` and updates the rolling data window
+    /// accordingly.
+    fn set_chunk(&mut self, chunk: &[u8]) {
+        // Remove bytes exceeding the window size. Start with the
+        // oldest bytes, which are at the beginning of the buffer.
+        // The buffer length is then the position where the chunk
+        // we've just received starts
+        self.window_and_chunk.drain(
+            ..self
+                .window_and_chunk
+                .len()
+                .saturating_sub(ZOPFLI_WINDOW_SIZE),
+        );
+        self.chunk_start = self.window_and_chunk.len();
+
+        self.window_and_chunk.extend_from_slice(chunk);
+
+        self.have_chunk = true;
+    }
+
+    /// Encodes the last chunk and finishes any partial bits.
+    /// The encoder will be unusable for further compression
+    /// after this method returns. This is intended to be an
+    /// implementation detail of the `Drop` trait and
+    /// [`finish`](Self::finish) method.
+    fn _finish(&mut self) -> Result<Option<W>, Error> {
+        if self.bitwise_writer.is_none() {
+            return Ok(None);
+        }
+
+        self.compress_chunk(true)?;
+
+        let mut bitwise_writer = self.bitwise_writer.take().unwrap();
+        bitwise_writer.finish_partial_bits()?;
+
+        Ok(Some(bitwise_writer.out))
+    }
 }
 
-/// Deflate a part, to allow deflate() to use multiple master blocks if
-/// needed.
+impl<W: Write> Write for DeflateEncoder<W> {
+    fn write(&mut self, buf: &[u8]) -> Result<usize, Error> {
+        // Any previous chunk is known to be non-last at this point,
+        // so compress it now
+        if self.have_chunk {
+            self.compress_chunk(false)?;
+        }
+
+        // Set the chunk to be used for the next compression operation
+        // to this chunk. We don't know whether it's last or not yet
+        self.set_chunk(buf);
+
+        Ok(buf.len())
+    }
+
+    fn flush(&mut self) -> Result<(), Error> {
+        self.bitwise_writer.as_mut().unwrap().out.flush()
+    }
+}
+
+impl<W: Write> Drop for DeflateEncoder<W> {
+    fn drop(&mut self) {
+        self._finish().ok();
+    }
+}
+
+/// Deflate a part, to allow for chunked, streaming compression with [`DeflateEncoder`].
 /// It is possible to call this function multiple times in a row, shifting
 /// instart and inend to next bytes of the data. If instart is larger than 0, then
 /// previous bytes are used as the initial dictionary for LZ77.
@@ -170,7 +170,7 @@ where
 /// Like deflate, but allows to specify start and end byte with instart and
 /// inend. Only that part is compressed, but earlier bytes are still used for the
 /// back window.
-fn deflate_part<W>(
+fn deflate_part<W: Write>(
     options: &Options,
     btype: BlockType,
     final_block: bool,
@@ -178,10 +178,7 @@ fn deflate_part<W>(
     instart: usize,
     inend: usize,
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<()>
-where
-    W: Write,
-{
+) -> Result<(), Error> {
     /* If btype=Dynamic is specified, it tries all block types. If a lesser btype is
     given, then however it forces that one. Neither of the lesser types needs
     block splitting as they have no dynamic huffman trees. */
@@ -191,9 +188,14 @@ where
         }
         BlockType::Fixed => {
             let mut store = Lz77Store::new();
-            let mut s = ZopfliBlockState::new(options, instart, inend);
 
-            lz77_optimal_fixed(&mut s, in_data, instart, inend, &mut store);
+            lz77_optimal_fixed(
+                &mut ZopfliLongestMatchCache::new(inend - instart),
+                in_data,
+                instart,
+                inend,
+                &mut store,
+            );
             add_lz77_block(
                 btype,
                 final_block,
@@ -216,13 +218,39 @@ where
     }
 }
 
+/// The type of data blocks to generate for a DEFLATE stream.
 #[derive(PartialEq, Eq, Copy, Clone, Debug)]
+#[cfg_attr(all(test, feature = "std"), derive(proptest_derive::Arbitrary))]
 pub enum BlockType {
+    /// Non-compressed blocks (BTYPE=00).
+    ///
+    /// The input data will be divided into chunks up to 64 KiB big and
+    /// stored in the DEFLATE stream without compression. This is mainly
+    /// useful for test and development purposes.
     Uncompressed,
+    /// Compressed blocks with fixed Huffman codes (BTYPE=01).
+    ///
+    /// The input data will be compressed into DEFLATE blocks using a fixed
+    /// Huffman tree defined in the DEFLATE specification. This provides fast
+    /// but poor compression, as the Zopfli algorithm is not actually used.
     Fixed,
+    /// Select the most space-efficient block types for the input data.
+    /// This is the recommended type for the vast majority of Zopfli
+    /// applications.
+    ///
+    /// This mode lets the Zopfli algorithm choose the combination of block
+    /// types that minimizes data size. The emitted block types may be
+    /// [`Uncompressed`](Self::Uncompressed) or [`Fixed`](Self::Fixed), in
+    /// addition to compressed with dynamic Huffman codes (BTYPE=10).
     Dynamic,
 }
 
+impl Default for BlockType {
+    fn default() -> Self {
+        Self::Dynamic
+    }
+}
+
 fn fixed_tree() -> (Vec<u32>, Vec<u32>) {
     let mut ll = Vec::with_capacity(ZOPFLI_NUM_LL);
     ll.resize(144, 8);
@@ -363,11 +391,11 @@ fn calculate_block_symbol_size_small(
             LitLen::LengthDist(litlens_i, dists_i) => {
                 debug_assert!(litlens_i < 259);
                 let ll_symbol = get_length_symbol(litlens_i as usize);
-                let d_symbol = get_dist_symbol(dists_i as i32);
-                result += ll_lengths[ll_symbol as usize];
-                result += d_lengths[d_symbol as usize];
-                result += get_length_symbol_extra_bits(ll_symbol) as u32;
-                result += get_dist_symbol_extra_bits(d_symbol) as u32;
+                let d_symbol = get_dist_symbol(dists_i);
+                result += ll_lengths[ll_symbol];
+                result += d_lengths[d_symbol];
+                result += get_length_symbol_extra_bits(ll_symbol);
+                result += get_dist_symbol_extra_bits(d_symbol);
             }
         }
     }
@@ -394,11 +422,11 @@ fn calculate_block_symbol_size_given_counts(
         }
         for i in 257..286 {
             result += ll_lengths[i] * ll_counts[i] as u32;
-            result += (get_length_symbol_extra_bits(i as i32) * ll_counts[i] as i32) as u32;
+            result += (get_length_symbol_extra_bits(i) as usize * ll_counts[i]) as u32;
         }
         for i in 0..30 {
             result += d_lengths[i] * d_counts[i] as u32;
-            result += (get_dist_symbol_extra_bits(i as i32) * d_counts[i] as i32) as u32;
+            result += (get_dist_symbol_extra_bits(i) as usize * d_counts[i]) as u32;
         }
         result += ll_lengths[256]; // end symbol
         result as usize
@@ -418,7 +446,13 @@ fn calculate_block_symbol_size(
     } else {
         let (ll_counts, d_counts) = lz77.get_histogram(lstart, lend);
         calculate_block_symbol_size_given_counts(
-            &ll_counts, &d_counts, ll_lengths, d_lengths, lz77, lstart, lend,
+            &*ll_counts,
+            &*d_counts,
+            ll_lengths,
+            d_lengths,
+            lz77,
+            lstart,
+            lend,
         )
     }
 }
@@ -567,17 +601,14 @@ fn calculate_tree_size(ll_lengths: &[u32], d_lengths: &[u32]) -> usize {
 
 /// Encodes the Huffman tree and returns how many bits its encoding takes and returns output.
 // TODO: This return value is unused.
-fn encode_tree<W>(
+fn encode_tree<W: Write>(
     ll_lengths: &[u32],
     d_lengths: &[u32],
     use_16: bool,
     use_17: bool,
     use_18: bool,
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<usize>
-where
-    W: Write,
-{
+) -> Result<usize, Error> {
     let mut hlit = 29; /* 286 - 257 */
     let mut hdist = 29; /* 32 - 1, but gzip does not like hdist > 29.*/
 
@@ -727,14 +758,11 @@ where
     Ok(result_size)
 }
 
-fn add_dynamic_tree<W>(
+fn add_dynamic_tree<W: Write>(
     ll_lengths: &[u32],
     d_lengths: &[u32],
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<()>
-where
-    W: Write,
-{
+) -> Result<(), Error> {
     let mut best = 0;
     let mut bestsize = 0;
 
@@ -771,7 +799,7 @@ where
 ///   set it to `0` to not do the assertion.
 /// `bitwise_writer`: writer responsible for appending bits
 #[allow(clippy::too_many_arguments)] // Not feasible to refactor in a more readable way
-fn add_lz77_block<W>(
+fn add_lz77_block<W: Write>(
     btype: BlockType,
     final_block: bool,
     in_data: &[u8],
@@ -780,10 +808,7 @@ fn add_lz77_block<W>(
     lend: usize,
     expected_data_size: usize,
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<()>
-where
-    W: Write,
-{
+) -> Result<(), Error> {
     if btype == BlockType::Uncompressed {
         let length = lz77.get_byte_range(lstart, lend);
         let pos = if lstart == lend { 0 } else { lz77.pos[lstart] };
@@ -892,8 +917,8 @@ fn try_optimize_huffman_for_rle(
     ll_lengths: Vec<u32>,
     d_lengths: Vec<u32>,
 ) -> (f64, Vec<u32>, Vec<u32>) {
-    let mut ll_counts2 = ll_counts.to_owned();
-    let mut d_counts2 = d_counts.to_owned();
+    let mut ll_counts2 = Vec::from(ll_counts);
+    let mut d_counts2 = Vec::from(d_counts);
 
     let treesize = calculate_tree_size(&ll_lengths, &d_lengths);
     let datasize = calculate_block_symbol_size_given_counts(
@@ -940,13 +965,19 @@ fn get_dynamic_lengths(lz77: &Lz77Store, lstart: usize, lend: usize) -> (f64, Ve
     let (mut ll_counts, d_counts) = lz77.get_histogram(lstart, lend);
     ll_counts[256] = 1; /* End symbol. */
 
-    let ll_lengths = length_limited_code_lengths(&ll_counts, 15);
-    let mut d_lengths = length_limited_code_lengths(&d_counts, 15);
+    let ll_lengths = length_limited_code_lengths(&*ll_counts, 15);
+    let mut d_lengths = length_limited_code_lengths(&*d_counts, 15);
 
     patch_distance_codes_for_buggy_decoders(&mut d_lengths[..]);
 
     try_optimize_huffman_for_rle(
-        lz77, lstart, lend, &ll_counts, &d_counts, ll_lengths, d_lengths,
+        lz77,
+        lstart,
+        lend,
+        &*ll_counts,
+        &*d_counts,
+        ll_lengths,
+        d_lengths,
     )
 }
 
@@ -954,7 +985,7 @@ fn get_dynamic_lengths(lz77: &Lz77Store, lstart: usize, lend: usize) -> (f64, Ve
 /// end code 256. `expected_data_size` is the uncompressed block size, used for
 /// assert, but you can set it to `0` to not do the assertion.
 #[allow(clippy::too_many_arguments)] // Not feasible to refactor in a more readable way
-fn add_lz77_data<W>(
+fn add_lz77_data<W: Write>(
     lz77: &Lz77Store,
     lstart: usize,
     lend: usize,
@@ -964,10 +995,7 @@ fn add_lz77_data<W>(
     d_symbols: &[u32],
     d_lengths: &[u32],
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<()>
-where
-    W: Write,
-{
+) -> Result<(), Error> {
     let mut testlength = 0;
 
     for &item in &lz77.litlens[lstart..lend] {
@@ -981,21 +1009,20 @@ where
             }
             LitLen::LengthDist(len, dist) => {
                 let litlen = len as usize;
-                let lls = get_length_symbol(litlen) as u32;
-                let ds = get_dist_symbol(dist as i32) as u32;
+                let lls = get_length_symbol(litlen);
+                let ds = get_dist_symbol(dist);
                 debug_assert!((3..=288).contains(&litlen));
-                debug_assert!(ll_lengths[lls as usize] > 0);
-                debug_assert!(d_lengths[ds as usize] > 0);
-                bitwise_writer
-                    .add_huffman_bits(ll_symbols[lls as usize], ll_lengths[lls as usize])?;
+                debug_assert!(ll_lengths[lls] > 0);
+                debug_assert!(d_lengths[ds] > 0);
+                bitwise_writer.add_huffman_bits(ll_symbols[lls], ll_lengths[lls])?;
                 bitwise_writer.add_bits(
-                    get_length_extra_bits_value(litlen as i32) as u32,
+                    get_length_extra_bits_value(litlen),
                     get_length_extra_bits(litlen) as u32,
                 )?;
-                bitwise_writer.add_huffman_bits(d_symbols[ds as usize], d_lengths[ds as usize])?;
+                bitwise_writer.add_huffman_bits(d_symbols[ds], d_lengths[ds])?;
                 bitwise_writer.add_bits(
-                    get_dist_extra_bits_value(dist as i32) as u32,
-                    get_dist_extra_bits(dist as i32) as u32,
+                    get_dist_extra_bits_value(dist) as u32,
+                    get_dist_extra_bits(dist) as u32,
                 )?;
                 testlength += litlen;
             }
@@ -1006,8 +1033,7 @@ where
 }
 
 #[allow(clippy::too_many_arguments)] // Not feasible to refactor in a more readable way
-fn add_lz77_block_auto_type<W>(
-    options: &Options,
+fn add_lz77_block_auto_type<W: Write>(
     final_block: bool,
     in_data: &[u8],
     lz77: &Lz77Store,
@@ -1015,10 +1041,7 @@ fn add_lz77_block_auto_type<W>(
     lend: usize,
     expected_data_size: usize,
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<()>
-where
-    W: Write,
-{
+) -> Result<(), Error> {
     let uncompressedcost = calculate_block_size(lz77, lstart, lend, BlockType::Uncompressed);
     let mut fixedcost = calculate_block_size(lz77, lstart, lend, BlockType::Fixed);
     let dyncost = calculate_block_size(lz77, lstart, lend, BlockType::Dynamic);
@@ -1041,12 +1064,17 @@ where
         let instart = lz77.pos[lstart];
         let inend = instart + lz77.get_byte_range(lstart, lend);
 
-        let mut s = ZopfliBlockState::new(options, instart, inend);
-        lz77_optimal_fixed(&mut s, in_data, instart, inend, &mut fixedstore);
+        lz77_optimal_fixed(
+            &mut ZopfliLongestMatchCache::new(inend - instart),
+            in_data,
+            instart,
+            inend,
+            &mut fixedstore,
+        );
         fixedcost = calculate_block_size(&fixedstore, 0, fixedstore.size(), BlockType::Fixed);
     }
 
-    if uncompressedcost < fixedcost && uncompressedcost < dyncost {
+    if uncompressedcost <= fixedcost && uncompressedcost <= dyncost {
         add_lz77_block(
             BlockType::Uncompressed,
             final_block,
@@ -1057,7 +1085,7 @@ where
             expected_data_size,
             bitwise_writer,
         )
-    } else if fixedcost < dyncost {
+    } else if fixedcost <= dyncost {
         if expensivefixed {
             add_lz77_block(
                 BlockType::Fixed,
@@ -1109,24 +1137,19 @@ pub fn calculate_block_size_auto_type(lz77: &Lz77Store, lstart: usize, lend: usi
     uncompressedcost.min(fixedcost).min(dyncost)
 }
 
-fn add_all_blocks<W>(
+fn add_all_blocks<W: Write>(
     splitpoints: &[usize],
     lz77: &Lz77Store,
-    options: &Options,
     final_block: bool,
     in_data: &[u8],
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<()>
-where
-    W: Write,
-{
+) -> Result<(), Error> {
     let mut last = 0;
     for &item in splitpoints.iter() {
-        add_lz77_block_auto_type(options, false, in_data, lz77, last, item, 0, bitwise_writer)?;
+        add_lz77_block_auto_type(false, in_data, lz77, last, item, 0, bitwise_writer)?;
         last = item;
     }
     add_lz77_block_auto_type(
-        options,
         final_block,
         in_data,
         lz77,
@@ -1137,17 +1160,14 @@ where
     )
 }
 
-fn blocksplit_attempt<W>(
+fn blocksplit_attempt<W: Write>(
     options: &Options,
     final_block: bool,
     in_data: &[u8],
     instart: usize,
     inend: usize,
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<()>
-where
-    W: Write,
-{
+) -> Result<(), Error> {
     let mut totalcost = 0.0;
     let mut lz77 = Lz77Store::new();
 
@@ -1155,7 +1175,6 @@ where
     let mut splitpoints_uncompressed = Vec::with_capacity(options.maximum_block_splits as usize);
 
     blocksplit(
-        options,
         in_data,
         instart,
         inend,
@@ -1167,9 +1186,14 @@ where
 
     let mut last = instart;
     for &item in &splitpoints_uncompressed {
-        let mut s = ZopfliBlockState::new(options, last, item);
-
-        let store = lz77_optimal(&mut s, in_data, last, item, options.iteration_count.get());
+        let store = lz77_optimal(
+            &mut ZopfliLongestMatchCache::new(item - last),
+            in_data,
+            last,
+            item,
+            options.iteration_count.get(),
+            options.iterations_without_improvement.get(),
+        );
         totalcost += calculate_block_size_auto_type(&store, 0, store.size());
 
         // ZopfliAppendLZ77Store(&store, &lz77);
@@ -1183,9 +1207,14 @@ where
         last = item;
     }
 
-    let mut s = ZopfliBlockState::new(options, last, inend);
-
-    let store = lz77_optimal(&mut s, in_data, last, inend, options.iteration_count.get());
+    let store = lz77_optimal(
+        &mut ZopfliLongestMatchCache::new(inend - last),
+        in_data,
+        last,
+        inend,
+        options.iteration_count.get(),
+        options.iterations_without_improvement.get(),
+    );
     totalcost += calculate_block_size_auto_type(&store, 0, store.size());
 
     // ZopfliAppendLZ77Store(&store, &lz77);
@@ -1213,31 +1242,33 @@ where
         }
     }
 
-    add_all_blocks(
-        &splitpoints,
-        &lz77,
-        options,
-        final_block,
-        in_data,
-        bitwise_writer,
-    )
+    add_all_blocks(&splitpoints, &lz77, final_block, in_data, bitwise_writer)
 }
 
 /// Since an uncompressed block can be max 65535 in size, it actually adds
-/// multible blocks if needed.
-fn add_non_compressed_block<W>(
+/// multiple blocks if needed.
+fn add_non_compressed_block<W: Write>(
     final_block: bool,
     in_data: &[u8],
     instart: usize,
     inend: usize,
     bitwise_writer: &mut BitwiseWriter<W>,
-) -> io::Result<()>
-where
-    W: Write,
-{
+) -> Result<(), Error> {
     let in_data = &in_data[instart..inend];
 
-    for (chunk, is_final) in in_data.chunks(65535).flag_last() {
+    let in_data_chunks = in_data.chunks(65535).size_hint().0;
+
+    for (chunk, is_final) in in_data
+        .chunks(65535)
+        .flag_last()
+        // Make sure that we output at least one chunk if this is the final block
+        .chain(iter::once((&[][..], true)))
+        .take(if final_block {
+            cmp::max(in_data_chunks, 1)
+        } else {
+            in_data_chunks
+        })
+    {
         let blocksize = chunk.len();
         let nlen = !blocksize;
 
@@ -1259,17 +1290,14 @@ where
     Ok(())
 }
 
-pub struct BitwiseWriter<W> {
+struct BitwiseWriter<W> {
     bit: u8,
     bp: u8,
     len: usize,
     out: W,
 }
 
-impl<W> BitwiseWriter<W>
-where
-    W: Write,
-{
+impl<W: Write> BitwiseWriter<W> {
     fn new(out: W) -> BitwiseWriter<W> {
         BitwiseWriter {
             bit: 0,
@@ -1284,17 +1312,17 @@ where
     }
 
     /// For when you want to add a full byte.
-    fn add_byte(&mut self, byte: u8) -> io::Result<()> {
+    fn add_byte(&mut self, byte: u8) -> Result<(), Error> {
         self.add_bytes(&[byte])
     }
 
     /// For adding a slice of bytes.
-    fn add_bytes(&mut self, bytes: &[u8]) -> io::Result<()> {
+    fn add_bytes(&mut self, bytes: &[u8]) -> Result<(), Error> {
         self.len += bytes.len();
         self.out.write_all(bytes)
     }
 
-    fn add_bit(&mut self, bit: u8) -> io::Result<()> {
+    fn add_bit(&mut self, bit: u8) -> Result<(), Error> {
         self.bit |= bit << self.bp;
         self.bp += 1;
         if self.bp == 8 {
@@ -1304,7 +1332,7 @@ where
         }
     }
 
-    fn add_bits(&mut self, symbol: u32, length: u32) -> io::Result<()> {
+    fn add_bits(&mut self, symbol: u32, length: u32) -> Result<(), Error> {
         // TODO: make more efficient (add more bits at once)
         for i in 0..length {
             let bit = ((symbol >> i) & 1) as u8;
@@ -1316,7 +1344,7 @@ where
 
     /// Adds bits, like `add_bits`, but the order is inverted. The deflate specification
     /// uses both orders in one standard.
-    fn add_huffman_bits(&mut self, symbol: u32, length: u32) -> io::Result<()> {
+    fn add_huffman_bits(&mut self, symbol: u32, length: u32) -> Result<(), Error> {
         // TODO: make more efficient (add more bits at once)
         for i in 0..length {
             let bit = ((symbol >> (length - i - 1)) & 1) as u8;
@@ -1326,7 +1354,7 @@ where
         Ok(())
     }
 
-    fn finish_partial_bits(&mut self) -> io::Result<()> {
+    fn finish_partial_bits(&mut self) -> Result<(), Error> {
         if self.bp != 0 {
             let bytes = &[self.bit];
             self.add_bytes(bytes)?;
@@ -1345,6 +1373,8 @@ fn set_counts_to_count(counts: &mut [usize], count: usize, i: usize, stride: usi
 
 #[cfg(test)]
 mod test {
+    use miniz_oxide::inflate;
+
     use super::*;
 
     #[test]
@@ -1358,4 +1388,32 @@ mod test {
 
         assert_eq!(counts, vec![0, 1, 2, 100, 100, 100, 100, 100, 8, 9])
     }
+
+    #[test]
+    fn weird_encoder_write_size_combinations_works() {
+        let mut compressed_data = vec![];
+
+        let default_options = Options::default();
+        let mut encoder =
+            DeflateEncoder::new(default_options, BlockType::default(), &mut compressed_data);
+
+        encoder.write_all(&[0]).unwrap();
+        encoder.write_all(&[]).unwrap();
+        encoder.write_all(&[1, 2]).unwrap();
+        encoder.write_all(&[]).unwrap();
+        encoder.write_all(&[]).unwrap();
+        encoder.write_all(&[3]).unwrap();
+        encoder.write_all(&[4]).unwrap();
+
+        encoder.finish().unwrap();
+
+        let decompressed_data = inflate::decompress_to_vec(&compressed_data)
+            .expect("Could not inflate compressed stream");
+
+        assert_eq!(
+            &[0, 1, 2, 3, 4][..],
+            decompressed_data,
+            "Decompressed data should match input data"
+        );
+    }
 }
