@@ -1,58 +1,39 @@
+use alloc::{
+    alloc::{alloc, handle_alloc_error, Layout},
+    boxed::Box,
+};
+use core::ptr::{addr_of, addr_of_mut, NonNull};
+
 use crate::util::{ZOPFLI_MIN_MATCH, ZOPFLI_WINDOW_MASK, ZOPFLI_WINDOW_SIZE};
 
 const HASH_SHIFT: i32 = 5;
 const HASH_MASK: u16 = 32767;
 
-#[derive(Debug, Copy, Clone, PartialEq, Eq)]
+#[derive(Copy, Clone, PartialEq, Eq)]
 pub enum Which {
     Hash1,
     Hash2,
 }
 
+#[derive(Clone)]
 pub struct SmallerHashThing {
     prev: u16,            /* Index to index of prev. occurrence of same hash. */
     hashval: Option<u16>, /* Index to hash value at this index. */
 }
 
+#[derive(Clone)]
 pub struct HashThing {
-    head: Vec<i32>, /* Hash value to index of its most recent occurrence. */
-    prev_and_hashval: Vec<SmallerHashThing>,
+    head: [i16; 65536], /* Hash value to index of its most recent occurrence. */
+    prev_and_hashval: [SmallerHashThing; ZOPFLI_WINDOW_SIZE],
     val: u16, /* Current hash value. */
 }
 
 impl HashThing {
-    fn new() -> HashThing {
-        let mut prev_and_hashval = Vec::with_capacity(ZOPFLI_WINDOW_SIZE);
-        for p in 0..ZOPFLI_WINDOW_SIZE as u16 {
-            prev_and_hashval.push(SmallerHashThing {
-                prev: p,
-                hashval: None,
-            });
-        }
-        HashThing {
-            head: vec![-1; 65536],
-            prev_and_hashval,
-            val: 0,
-        }
-    }
-
-    fn reset(&mut self) {
-        self.val = 0;
-        self.head.clear();
-        self.head.resize(65536, -1);
-        self.prev_and_hashval = (0..ZOPFLI_WINDOW_SIZE as u16)
-            .map(|p| SmallerHashThing {
-                prev: p,
-                hashval: None,
-            })
-            .collect();
-    }
-
     fn update(&mut self, hpos: usize) {
         let hashval = self.val;
         let index = self.val as usize;
         let head_index = self.head[index];
-        let prev = if head_index != -1
+        let prev = if head_index >= 0
             && self.prev_and_hashval[head_index as usize]
                 .hashval
                 .map_or(false, |hv| hv == self.val)
@@ -66,10 +47,11 @@ impl HashThing {
             prev,
             hashval: Some(hashval),
         };
-        self.head[index] = hpos as i32;
+        self.head[index] = hpos as i16;
     }
 }
 
+#[derive(Clone)]
 pub struct ZopfliHash {
     hash1: HashThing,
     hash2: HashThing,
@@ -77,18 +59,59 @@ pub struct ZopfliHash {
 }
 
 impl ZopfliHash {
-    pub fn new() -> ZopfliHash {
-        ZopfliHash {
-            hash1: HashThing::new(),
-            hash2: HashThing::new(),
-            same: [0; ZOPFLI_WINDOW_SIZE],
+    pub fn new() -> Box<ZopfliHash> {
+        const LAYOUT: Layout = Layout::new::<ZopfliHash>();
+
+        let ptr = NonNull::new(unsafe { alloc(LAYOUT) } as *mut ZopfliHash)
+            .unwrap_or_else(|| handle_alloc_error(LAYOUT));
+
+        unsafe {
+            Self::init(ptr);
+            Box::from_raw(ptr.as_ptr())
         }
     }
 
+    /// Initializes the [`ZopfliHash`] instance pointed by `hash` to an initial state.
+    ///
+    /// ## Safety
+    /// `hash` must point to aligned, valid memory for writes.
+    unsafe fn init(hash: NonNull<Self>) {
+        let hash = hash.as_ptr();
+
+        // SAFETY: addr_of(_mut) macros are used to avoid creating intermediate references, which
+        //         are undefined behavior when data is uninitialized. Note that it also is UB to
+        //         assume that integer values and arrays can be read after allocating their memory:
+        //         the allocator returns valid, but uninitialized pointers that are not guaranteed
+        //         to hold a fixed bit pattern (c.f. core::mem::MaybeUnit docs and
+        //         https://doc.rust-lang.org/std/ptr/index.html#safety).
+
+        for i in 0..ZOPFLI_WINDOW_SIZE {
+            // Arrays are guaranteed to be laid out with their elements placed in consecutive
+            // memory positions: https://doc.rust-lang.org/reference/type-layout.html#array-layout.
+            // Therefore, a pointer to an array has the same address as the pointer to its first
+            // element, and adding size_of::<N>() bytes to that address yields the address of the
+            // second element, and so on.
+            let prev_and_hashval =
+                (addr_of_mut!((*hash).hash1.prev_and_hashval) as *mut SmallerHashThing).add(i);
+            addr_of_mut!((*prev_and_hashval).prev).write(i as u16);
+            addr_of_mut!((*prev_and_hashval).hashval).write(None);
+        }
+
+        // Rust signed integers are guaranteed to be represented in two's complement notation:
+        // https://doc.rust-lang.org/reference/types/numeric.html#integer-types
+        // In this notation, -1 is expressed as an all-ones value. Therefore, writing
+        // size_of::<[i16; N]> all-ones bytes initializes all of them to -1.
+        addr_of_mut!((*hash).hash1.head).write_bytes(0xFF, 1);
+        addr_of_mut!((*hash).hash1.val).write(0);
+
+        addr_of_mut!((*hash).hash2).copy_from_nonoverlapping(addr_of!((*hash).hash1), 1);
+
+        // Zero-initializes all the array elements
+        addr_of_mut!((*hash).same).write_bytes(0, 1);
+    }
+
     pub fn reset(&mut self) {
-        self.hash1.reset();
-        self.hash2.reset();
-        self.same = [0; ZOPFLI_WINDOW_SIZE];
+        unsafe { Self::init(NonNull::new(self).unwrap()) }
     }
 
     pub fn warmup(&mut self, arr: &[u8], pos: usize, end: usize) {
@@ -131,18 +154,11 @@ impl ZopfliHash {
         self.hash2.update(hpos);
     }
 
-    pub fn head_at(&self, index: usize, which: Which) -> i32 {
-        match which {
-            Which::Hash1 => self.hash1.head[index],
-            Which::Hash2 => self.hash2.head[index],
-        }
-    }
-
-    pub fn prev_at(&self, index: usize, which: Which) -> u16 {
-        match which {
+    pub fn prev_at(&self, index: usize, which: Which) -> usize {
+        (match which {
             Which::Hash1 => self.hash1.prev_and_hashval[index].prev,
             Which::Hash2 => self.hash2.prev_and_hashval[index].prev,
-        }
+        }) as usize
     }
 
     pub fn hash_val_at(&self, index: usize, which: Which) -> i32 {
