@@ -1,15 +1,17 @@
-use std::cmp;
+use alloc::{boxed::Box, vec::Vec};
+use core::cmp;
 
-use crate::cache::{Cache, NoCache, ZopfliLongestMatchCache};
-use crate::hash::{Which, ZopfliHash};
-use crate::symbols::{get_dist_symbol, get_length_symbol};
-use crate::util::{
-    ZOPFLI_MAX_CHAIN_HITS, ZOPFLI_MAX_MATCH, ZOPFLI_MIN_MATCH, ZOPFLI_NUM_D, ZOPFLI_NUM_LL,
-    ZOPFLI_WINDOW_MASK, ZOPFLI_WINDOW_SIZE,
+use crate::{
+    cache::Cache,
+    hash::{Which, ZopfliHash},
+    symbols::{get_dist_symbol, get_length_symbol},
+    util::{
+        boxed_array, ZOPFLI_MAX_CHAIN_HITS, ZOPFLI_MAX_MATCH, ZOPFLI_MIN_MATCH, ZOPFLI_NUM_D,
+        ZOPFLI_NUM_LL, ZOPFLI_WINDOW_MASK, ZOPFLI_WINDOW_SIZE,
+    },
 };
-use crate::Options;
 
-#[derive(Clone, Debug, Copy)]
+#[derive(Clone, Copy)]
 pub enum LitLen {
     Literal(u16),
     LengthDist(u16, u16),
@@ -29,7 +31,7 @@ impl LitLen {
 /// Parameter dists: Contains the distances. A value is 0 to indicate that there is
 /// no dist and the corresponding litlens value is a literal instead of a length.
 /// Parameter size: The size of both the litlens and dists arrays.
-#[derive(Debug, Clone, Default)]
+#[derive(Clone, Default)]
 pub struct Lz77Store {
     pub litlens: Vec<LitLen>,
 
@@ -79,9 +81,9 @@ impl Lz77Store {
             if origsize == 0 {
                 self.ll_counts.resize(origsize + ZOPFLI_NUM_LL, 0);
             } else {
-                let mut last_histogram =
-                    self.ll_counts[(origsize - ZOPFLI_NUM_LL)..origsize].to_vec();
-                self.ll_counts.append(&mut last_histogram);
+                // Append last histogram
+                self.ll_counts
+                    .extend_from_within((origsize - ZOPFLI_NUM_LL)..origsize);
             }
         }
 
@@ -89,9 +91,9 @@ impl Lz77Store {
             if origsize == 0 {
                 self.d_counts.resize(ZOPFLI_NUM_D, 0);
             } else {
-                let mut last_histogram =
-                    self.d_counts[(origsize - ZOPFLI_NUM_D)..origsize].to_vec();
-                self.d_counts.append(&mut last_histogram);
+                // Append last histogram
+                self.d_counts
+                    .extend_from_within((origsize - ZOPFLI_NUM_D)..origsize);
             }
         }
 
@@ -110,9 +112,9 @@ impl Lz77Store {
             LitLen::LengthDist(length, dist) => {
                 let len_sym = get_length_symbol(length as usize);
                 self.ll_symbol.push(len_sym as u16);
-                self.d_symbol.push(get_dist_symbol(dist as i32) as u16);
-                self.ll_counts[llstart + len_sym as usize] += 1;
-                self.d_counts[dstart + get_dist_symbol(dist as i32) as usize] += 1;
+                self.d_symbol.push(get_dist_symbol(dist) as u16);
+                self.ll_counts[llstart + len_sym] += 1;
+                self.d_counts[dstart + get_dist_symbol(dist)] += 1;
             }
         }
     }
@@ -132,15 +134,7 @@ impl Lz77Store {
     /// The result is placed in the Lz77Store.
     /// If instart is larger than 0, it uses values before instart as starting
     /// dictionary.
-    pub fn greedy<C>(
-        &mut self,
-        s: &mut ZopfliBlockState<C>,
-        in_data: &[u8],
-        instart: usize,
-        inend: usize,
-    ) where
-        C: Cache,
-    {
+    pub fn greedy<C: Cache>(&mut self, lmc: &mut C, in_data: &[u8], instart: usize, inend: usize) {
         if instart == inend {
             return;
         }
@@ -168,7 +162,7 @@ impl Lz77Store {
             h.update(arr, i);
 
             let longest_match =
-                find_longest_match(s, &mut h, arr, i, inend, ZOPFLI_MAX_MATCH, &mut None);
+                find_longest_match(lmc, &h, arr, i, inend, instart, ZOPFLI_MAX_MATCH, &mut None);
             dist = longest_match.distance;
             leng = longest_match.length;
             lengthscore = get_length_score(leng as i32, dist as i32);
@@ -231,16 +225,14 @@ impl Lz77Store {
         }
     }
 
-    pub fn follow_path<C>(
+    pub fn follow_path<C: Cache>(
         &mut self,
         in_data: &[u8],
         instart: usize,
         inend: usize,
         path: Vec<u16>,
-        s: &mut ZopfliBlockState<C>,
-    ) where
-        C: Cache,
-    {
+        lmc: &mut C,
+    ) {
         let windowstart = instart.saturating_sub(ZOPFLI_WINDOW_SIZE);
 
         if instart == inend {
@@ -257,7 +249,7 @@ impl Lz77Store {
         }
 
         let mut pos = instart;
-        for &item in &path {
+        for item in path.into_iter().rev() {
             let mut length = item;
             debug_assert!(pos < inend);
 
@@ -267,8 +259,16 @@ impl Lz77Store {
             if length >= ZOPFLI_MIN_MATCH as u16 {
                 // Get the distance by recalculating longest match. The found length
                 // should match the length from the path.
-                let longest_match =
-                    find_longest_match(s, &mut h, arr, pos, inend, length as usize, &mut None);
+                let longest_match = find_longest_match(
+                    lmc,
+                    &h,
+                    arr,
+                    pos,
+                    inend,
+                    instart,
+                    length as usize,
+                    &mut None,
+                );
                 let dist = longest_match.distance;
                 let dummy_length = longest_match.length;
                 debug_assert!(!(dummy_length != length && length > 2 && dummy_length > 2));
@@ -288,9 +288,12 @@ impl Lz77Store {
         }
     }
 
-    fn get_histogram_at(&self, lpos: usize) -> (Vec<usize>, Vec<usize>) {
-        let mut ll = vec![0; ZOPFLI_NUM_LL];
-        let mut d = vec![0; ZOPFLI_NUM_D];
+    fn get_histogram_at(
+        &self,
+        lpos: usize,
+    ) -> (Box<[usize; ZOPFLI_NUM_LL]>, Box<[usize; ZOPFLI_NUM_D]>) {
+        let mut ll = boxed_array(0);
+        let mut d = boxed_array(0);
 
         /* The real histogram is created by using the histogram for this chunk, but
         all superfluous values of this chunk subtracted. */
@@ -321,10 +324,14 @@ impl Lz77Store {
     /// Gets the histogram of lit/len and dist symbols in the given range, using the
     /// cumulative histograms, so faster than adding one by one for large range. Does
     /// not add the one end symbol of value 256.
-    pub fn get_histogram(&self, lstart: usize, lend: usize) -> (Vec<usize>, Vec<usize>) {
+    pub fn get_histogram(
+        &self,
+        lstart: usize,
+        lend: usize,
+    ) -> (Box<[usize; ZOPFLI_NUM_LL]>, Box<[usize; ZOPFLI_NUM_D]>) {
         if lstart + ZOPFLI_NUM_LL * 3 > lend {
-            let mut ll_counts = vec![0; ZOPFLI_NUM_LL];
-            let mut d_counts = vec![0; ZOPFLI_NUM_D];
+            let mut ll_counts = boxed_array(0);
+            let mut d_counts = boxed_array(0);
             for i in lstart..lend {
                 ll_counts[self.ll_symbol[i] as usize] += 1;
                 if let LitLen::LengthDist(_, _) = self.litlens[i] {
@@ -344,11 +351,15 @@ impl Lz77Store {
                     ll.iter()
                         .zip(ll2.iter())
                         .map(|(&ll_item1, &ll_item2)| ll_item1 - ll_item2)
-                        .collect(),
+                        .collect::<Vec<_>>()
+                        .try_into()
+                        .unwrap(),
                     d.iter()
                         .zip(d2.iter())
                         .map(|(&d_item1, &d_item2)| d_item1 - d_item2)
-                        .collect(),
+                        .collect::<Vec<_>>()
+                        .try_into()
+                        .unwrap(),
                 )
             } else {
                 (ll, d)
@@ -366,72 +377,6 @@ impl Lz77Store {
     }
 }
 
-/// Some state information for compressing a block.
-/// This is currently a bit under-used (with mainly only the longest match cache),
-/// but is kept for easy future expansion.
-pub struct ZopfliBlockState<'a, C> {
-    pub options: &'a Options,
-    /* Cache for length/distance pairs found so far. */
-    lmc: C,
-    /* The start (inclusive) and end (not inclusive) of the current block. */
-    pub blockstart: usize,
-    pub blockend: usize,
-}
-
-impl<'a> ZopfliBlockState<'a, ZopfliLongestMatchCache> {
-    pub fn new(options: &'a Options, blockstart: usize, blockend: usize) -> Self {
-        ZopfliBlockState {
-            options,
-            blockstart,
-            blockend,
-            lmc: ZopfliLongestMatchCache::new(blockend - blockstart),
-        }
-    }
-}
-
-impl<'a> ZopfliBlockState<'a, NoCache> {
-    pub fn new_without_cache(options: &'a Options, blockstart: usize, blockend: usize) -> Self {
-        ZopfliBlockState {
-            options,
-            blockstart,
-            blockend,
-            lmc: NoCache,
-        }
-    }
-}
-
-impl<'a, C> ZopfliBlockState<'a, C>
-where
-    C: Cache,
-{
-    /// Gets distance, length and sublen values from the cache if possible.
-    /// Returns 1 if it got the values from the cache, 0 if not.
-    /// Updates the limit value to a smaller one if possible with more limited
-    /// information from the cache.
-    fn try_get_from_longest_match_cache(
-        &self,
-        pos: usize,
-        limit: usize,
-        sublen: &mut Option<&mut [u16]>,
-    ) -> LongestMatch {
-        self.lmc.try_get(pos, limit, sublen, self.blockstart)
-    }
-
-    /// Stores the found sublen, distance and length in the longest match cache, if
-    /// possible.
-    fn store_in_longest_match_cache(
-        &mut self,
-        pos: usize,
-        limit: usize,
-        sublen: &mut Option<&mut [u16]>,
-        distance: u16,
-        length: u16,
-    ) {
-        self.lmc
-            .store(pos, limit, sublen, distance, length, self.blockstart)
-    }
-}
-
 pub struct LongestMatch {
     pub distance: u16,
     pub length: u16,
@@ -475,19 +420,18 @@ fn get_match(array: &[u8], scan_offset: usize, match_offset: usize, end: usize)
     scan_offset
 }
 
-pub fn find_longest_match<C>(
-    s: &mut ZopfliBlockState<C>,
-    h: &mut ZopfliHash,
+#[allow(clippy::too_many_arguments)]
+pub fn find_longest_match<C: Cache>(
+    lmc: &mut C,
+    h: &ZopfliHash,
     array: &[u8],
     pos: usize,
     size: usize,
+    blockstart: usize,
     limit: usize,
     sublen: &mut Option<&mut [u16]>,
-) -> LongestMatch
-where
-    C: Cache,
-{
-    let mut longest_match = s.try_get_from_longest_match_cache(pos, limit, sublen);
+) -> LongestMatch {
+    let mut longest_match = lmc.try_get(pos, limit, sublen, blockstart);
 
     if longest_match.from_cache {
         debug_assert!(pos + (longest_match.length as usize) <= size);
@@ -516,37 +460,36 @@ where
 
     let (bestdist, bestlength) = find_longest_match_loop(h, array, pos, size, limit, sublen);
 
-    s.store_in_longest_match_cache(pos, limit, sublen, bestdist as u16, bestlength as u16);
+    lmc.store(pos, limit, sublen, bestdist, bestlength, blockstart);
 
-    debug_assert!(bestlength <= limit);
+    debug_assert!(bestlength <= limit as u16);
 
-    debug_assert!(pos + bestlength <= size);
-    longest_match.distance = bestdist as u16;
-    longest_match.length = bestlength as u16;
+    debug_assert!(pos + bestlength as usize <= size);
+    longest_match.distance = bestdist;
+    longest_match.length = bestlength;
     longest_match.from_cache = false;
     longest_match.limit = limit;
     longest_match
 }
 
 fn find_longest_match_loop(
-    h: &mut ZopfliHash,
+    h: &ZopfliHash,
     array: &[u8],
     pos: usize,
     size: usize,
     limit: usize,
     sublen: &mut Option<&mut [u16]>,
-) -> (i32, usize) {
+) -> (u16, u16) {
     let mut which_hash = Which::Hash1;
-    let mut pp = h.head_at(h.val(which_hash) as usize, which_hash); /* During the whole loop, p == hprev[pp]. */
-    let mut p = h.prev_at(pp as usize, which_hash);
-
     let hpos = pos & ZOPFLI_WINDOW_MASK;
-    debug_assert_eq!(pp as usize, hpos);
 
-    let mut dist = if (p as i32) < pp {
-        pp - (p as i32)
+    let mut pp = hpos; /* During the whole loop, p == hprev[pp]. */
+    let mut p = h.prev_at(pp, which_hash);
+
+    let mut dist = if p < pp {
+        pp - p
     } else {
-        (ZOPFLI_WINDOW_SIZE - (p as usize)) as i32 + pp
+        ZOPFLI_WINDOW_SIZE - p + pp
     };
 
     let mut bestlength = 1;
@@ -557,21 +500,18 @@ fn find_longest_match_loop(
     let mut match_offset;
 
     /* Go through all distances. */
-    while (dist as usize) < ZOPFLI_WINDOW_SIZE {
+    while dist < ZOPFLI_WINDOW_SIZE && chain_counter > 0 {
         let mut currentlength = 0;
 
-        debug_assert!((p as usize) < ZOPFLI_WINDOW_SIZE);
-        debug_assert_eq!(p, h.prev_at(pp as usize, which_hash));
-        debug_assert_eq!(
-            h.hash_val_at(p as usize, which_hash),
-            h.val(which_hash) as i32
-        );
+        debug_assert!(p < ZOPFLI_WINDOW_SIZE);
+        debug_assert_eq!(p, h.prev_at(pp, which_hash));
+        debug_assert_eq!(h.hash_val_at(p, which_hash), h.val(which_hash) as i32);
 
         if dist > 0 {
             debug_assert!(pos < size);
-            debug_assert!((dist as usize) <= pos);
+            debug_assert!(dist <= pos);
             scan_offset = pos;
-            match_offset = pos - (dist as usize);
+            match_offset = pos - dist;
 
             /* Testing the byte at position bestlength first, goes slightly faster. */
             if pos + bestlength >= size
@@ -579,12 +519,8 @@ fn find_longest_match_loop(
             {
                 let same0 = h.same[pos & ZOPFLI_WINDOW_MASK];
                 if same0 > 2 && array[scan_offset] == array[match_offset] {
-                    let same1 = h.same[(pos - (dist as usize)) & ZOPFLI_WINDOW_MASK];
-                    let same = [same0, same1, limit as u16]
-                        .iter()
-                        .min()
-                        .map(|x| *x as usize)
-                        .unwrap();
+                    let same1 = h.same[(pos - dist) & ZOPFLI_WINDOW_MASK];
+                    let same = cmp::min(cmp::min(same0, same1), limit as u16) as usize;
                     scan_offset += same;
                     match_offset += same;
                 }
@@ -609,30 +545,27 @@ fn find_longest_match_loop(
         /* Switch to the other hash once this will be more efficient. */
         if which_hash == Which::Hash1
             && bestlength >= h.same[hpos] as usize
-            && h.val(Which::Hash2) as i32 == h.hash_val_at(p as usize, Which::Hash2)
+            && h.val(Which::Hash2) as i32 == h.hash_val_at(p, Which::Hash2)
         {
             /* Now use the hash that encodes the length and first byte. */
             which_hash = Which::Hash2;
         }
 
-        pp = p as i32;
-        p = h.prev_at(p as usize, which_hash);
-        if (p as i32) == pp {
+        pp = p;
+        p = h.prev_at(p, which_hash);
+        if p == pp {
             break; /* Uninited prev value. */
         }
 
-        dist += if (p as i32) < pp {
-            pp - (p as i32)
+        dist += if p < pp {
+            pp - p
         } else {
-            (ZOPFLI_WINDOW_SIZE - (p as usize)) as i32 + pp
+            ZOPFLI_WINDOW_SIZE - p + pp
         };
 
         chain_counter -= 1;
-        if chain_counter == 0 {
-            break;
-        }
     }
-    (bestdist, bestlength)
+    (bestdist as u16, bestlength as u16)
 }
 
 /// Gets a score of the length given the distance. Typically, the score of the
@@ -664,13 +597,17 @@ fn get_length_score(length: i32, distance: i32) -> i32 {
     }
 }
 
+#[cfg(debug_assertions)]
 fn verify_len_dist(data: &[u8], pos: usize, dist: u16, length: u16) {
     for i in 0..length {
         let d1 = data[pos - (dist as usize) + (i as usize)];
         let d2 = data[pos + (i as usize)];
         if d1 != d2 {
-            debug_assert_eq!(d1, d2);
+            assert_eq!(d1, d2);
             break;
         }
     }
 }
+
+#[cfg(not(debug_assertions))]
+fn verify_len_dist(_data: &[u8], _pos: usize, _dist: u16, _length: u16) {}
