@@ -1,8 +1,10 @@
+use std::{
+    env,
+    fs::File,
+    io::{self, prelude::*},
+};
+
 use log::info;
-use std::env;
-use std::fs::File;
-use std::io::prelude::*;
-use std::io::{self, BufWriter};
 
 fn main() {
     let options = zopfli::Options::default();
@@ -27,9 +29,9 @@ fn main() {
         // Attempt to create the output file, panic if the output file could not be opened
         let out_file = File::create(&out_filename)
             .unwrap_or_else(|why| panic!("couldn't create output file {}: {}", out_filename, why));
-        let mut out_file = WriteStatistics::new(BufWriter::new(out_file));
+        let mut out_file = WriteStatistics::new(out_file);
 
-        zopfli::compress(&options, &output_type, &file, &mut out_file).unwrap_or_else(|why| {
+        zopfli::compress(options, output_type, &file, &mut out_file).unwrap_or_else(|why| {
             panic!("couldn't write to output file {}: {}", out_filename, why)
         });
 
@@ -54,10 +56,7 @@ impl<W> WriteStatistics<W> {
     }
 }
 
-impl<W> Write for WriteStatistics<W>
-where
-    W: Write,
-{
+impl<W: Write> Write for WriteStatistics<W> {
     fn write(&mut self, buf: &[u8]) -> io::Result<usize> {
         let res = self.inner.write(buf);
         if let Ok(size) = res {
