@@ -7,24 +7,28 @@
 //! multiple runs are done with updated cost models to converge to a better
 //! solution.
 
+use alloc::vec::Vec;
+use core::cmp;
+
 use log::{debug, trace};
-use std::{cmp, f64};
 
-use crate::cache::Cache;
-use crate::deflate::{calculate_block_size, BlockType};
-use crate::hash::ZopfliHash;
-use crate::lz77::{find_longest_match, LitLen, Lz77Store, ZopfliBlockState};
-use crate::symbols::{
-    get_dist_extra_bits, get_dist_symbol, get_length_extra_bits, get_length_symbol,
-};
-use crate::util::{
-    ZOPFLI_MAX_MATCH, ZOPFLI_NUM_D, ZOPFLI_NUM_LL, ZOPFLI_WINDOW_MASK, ZOPFLI_WINDOW_SIZE,
+use crate::{
+    cache::Cache,
+    deflate::{calculate_block_size, BlockType},
+    hash::ZopfliHash,
+    lz77::{find_longest_match, LitLen, Lz77Store},
+    symbols::{get_dist_extra_bits, get_dist_symbol, get_length_extra_bits, get_length_symbol},
+    util::{ZOPFLI_MAX_MATCH, ZOPFLI_NUM_D, ZOPFLI_NUM_LL, ZOPFLI_WINDOW_MASK, ZOPFLI_WINDOW_SIZE},
 };
 
-const K_INV_LOG2: f64 = f64::consts::LOG2_E; // 1.0 / log(2.0)
+const K_INV_LOG2: f64 = core::f64::consts::LOG2_E; // 1.0 / log(2.0)
+
+#[cfg(not(feature = "std"))]
+#[allow(unused_imports)] // False-positive
+use crate::math::F64MathExt;
 
 /// Cost model which should exactly match fixed tree.
-fn get_cost_fixed(litlen: u32, dist: u32) -> f64 {
+fn get_cost_fixed(litlen: usize, dist: u16) -> f64 {
     let result = if dist == 0 {
         if litlen <= 143 {
             8
@@ -32,30 +36,24 @@ fn get_cost_fixed(litlen: u32, dist: u32) -> f64 {
             9
         }
     } else {
-        let dbits = get_dist_extra_bits(dist as i32);
-        let lbits = get_length_extra_bits(litlen as usize);
-        let lsym = get_length_symbol(litlen as usize);
-        let mut cost = 0;
-        if lsym <= 279 {
-            cost += 7;
-        } else {
-            cost += 8;
-        }
-        cost += 5; // Every dist symbol has length 5.
-        cost + dbits + lbits
+        let dbits = get_dist_extra_bits(dist);
+        let lbits = get_length_extra_bits(litlen);
+        let lsym = get_length_symbol(litlen);
+        // Every dist symbol has length 5.
+        7 + (lsym > 279) as usize + 5 + dbits + lbits
     };
     result as f64
 }
 
 /// Cost model based on symbol statistics.
-fn get_cost_stat(litlen: u32, dist: u32, stats: &SymbolStats) -> f64 {
+fn get_cost_stat(litlen: usize, dist: u16, stats: &SymbolStats) -> f64 {
     if dist == 0 {
-        stats.ll_symbols[litlen as usize]
+        stats.ll_symbols[litlen]
     } else {
-        let lsym = get_length_symbol(litlen as usize) as usize;
-        let lbits = get_length_extra_bits(litlen as usize) as f64;
-        let dsym = get_dist_symbol(dist as i32) as usize;
-        let dbits = get_dist_extra_bits(dist as i32) as f64;
+        let lsym = get_length_symbol(litlen);
+        let lbits = get_length_extra_bits(litlen) as f64;
+        let dsym = get_dist_symbol(dist);
+        let dbits = get_dist_extra_bits(dist) as f64;
         lbits + dbits + stats.ll_symbols[lsym] + stats.d_symbols[dsym]
     }
 }
@@ -79,7 +77,7 @@ impl RanState {
     }
 }
 
-#[derive(Copy)]
+#[derive(Copy, Clone)]
 struct SymbolStats {
     /* The literal and length symbols. */
     litlens: [usize; ZOPFLI_NUM_LL],
@@ -92,12 +90,6 @@ struct SymbolStats {
     d_symbols: [f64; ZOPFLI_NUM_D],
 }
 
-impl Clone for SymbolStats {
-    fn clone(&self) -> Self {
-        *self
-    }
-}
-
 impl Default for SymbolStats {
     fn default() -> SymbolStats {
         SymbolStats {
@@ -174,8 +166,8 @@ impl SymbolStats {
             match litlen {
                 LitLen::Literal(lit) => self.litlens[lit as usize] += 1,
                 LitLen::LengthDist(len, dist) => {
-                    self.litlens[get_length_symbol(len as usize) as usize] += 1;
-                    self.dists[get_dist_symbol(dist as i32) as usize] += 1;
+                    self.litlens[get_length_symbol(len as usize)] += 1;
+                    self.dists[get_dist_symbol(dist)] += 1;
                 }
             }
         }
@@ -211,10 +203,7 @@ fn add_weighed_stat_freqs(
 
 /// Finds the minimum possible cost this cost model can return for valid length and
 /// distance symbols.
-fn get_cost_model_min_cost<F>(costmodel: F) -> f64
-where
-    F: Fn(u32, u32) -> f64,
-{
+fn get_cost_model_min_cost<F: Fn(usize, u16) -> f64>(costmodel: F) -> f64 {
     let mut bestlength = 0; // length that has lowest cost in the cost model
     let mut bestdist = 0; // distance that has lowest cost in the cost model
 
@@ -223,29 +212,29 @@ where
     // different symbols affect the cost model so only these need to be checked.
     // See RFC 1951 section 3.2.5. Compressed blocks (length and distance codes).
 
-    let dsymbols = [
+    const DSYMBOLS: [u16; 30] = [
         1, 2, 3, 4, 5, 7, 9, 13, 17, 25, 33, 49, 65, 97, 129, 193, 257, 385, 513, 769, 1025, 1537,
         2049, 3073, 4097, 6145, 8193, 12289, 16385, 24577,
     ];
 
-    let mut mincost = f64::MAX;
+    let mut mincost = f64::INFINITY;
     for i in 3..259 {
         let c = costmodel(i, 1);
         if c < mincost {
-            bestlength = i as i32;
+            bestlength = i;
             mincost = c;
         }
     }
 
-    mincost = f64::MAX;
-    for &dsym in dsymbols.iter().take(30) {
-        let c = costmodel(3, dsym as u32);
+    mincost = f64::INFINITY;
+    for dsym in DSYMBOLS {
+        let c = costmodel(3, dsym);
         if c < mincost {
             bestdist = dsym;
             mincost = c;
         }
     }
-    costmodel(bestlength as u32, bestdist as u32)
+    costmodel(bestlength, bestdist)
 }
 
 /// Performs the forward pass for "squeeze". Gets the most optimal length to reach
@@ -258,19 +247,15 @@ where
 /// `length_array`: output array of size `(inend - instart)` which will receive the best
 ///     length to reach this byte from a previous byte.
 /// returns the cost that was, according to the `costmodel`, needed to get to the end.
-fn get_best_lengths<F, C>(
-    s: &mut ZopfliBlockState<C>,
+fn get_best_lengths<F: Fn(usize, u16) -> f64, C: Cache>(
+    lmc: &mut C,
     in_data: &[u8],
     instart: usize,
     inend: usize,
     costmodel: F,
     h: &mut ZopfliHash,
     costs: &mut Vec<f32>,
-) -> (f64, Vec<u16>)
-where
-    F: Fn(u32, u32) -> f64,
-    C: Cache,
-{
+) -> (f64, Vec<u16>) {
     // Best cost to get here so far.
     let blocksize = inend - instart;
     let mut length_array = vec![0; blocksize + 1];
@@ -288,12 +273,10 @@ where
 
     costs.resize(blocksize + 1, 0.0);
     for cost in costs.iter_mut().take(blocksize + 1).skip(1) {
-        *cost = f32::MAX;
+        *cost = f32::INFINITY;
     }
     costs[0] = 0.0; /* Because it's the start. */
 
-    length_array[0] = 0;
-
     let mut i = instart;
     let mut leng;
     let mut longest_match;
@@ -310,7 +293,7 @@ where
             && i + ZOPFLI_MAX_MATCH * 2 + 1 < inend
             && h.same[(i - ZOPFLI_MAX_MATCH) & ZOPFLI_WINDOW_MASK] > ZOPFLI_MAX_MATCH as u16
         {
-            let symbolcost = costmodel(ZOPFLI_MAX_MATCH as u32, 1);
+            let symbolcost = costmodel(ZOPFLI_MAX_MATCH, 1);
             // Set the length to reach each one to ZOPFLI_MAX_MATCH, and the cost to
             // the cost corresponding to that length. Doing this, we skip
             // ZOPFLI_MAX_MATCH values to avoid calling ZopfliFindLongestMatch.
@@ -325,11 +308,12 @@ where
         }
 
         longest_match = find_longest_match(
-            s,
+            lmc,
             h,
             arr,
             i,
             inend,
+            instart,
             ZOPFLI_MAX_MATCH,
             &mut Some(&mut sublen),
         );
@@ -337,7 +321,7 @@ where
 
         // Literal.
         if i < inend {
-            let new_cost = costmodel(arr[i] as u32, 0) + costs[j] as f64;
+            let new_cost = costmodel(arr[i] as usize, 0) + costs[j] as f64;
             debug_assert!(new_cost >= 0.0);
             if new_cost < costs[j + 1] as f64 {
                 costs[j + 1] = new_cost as f32;
@@ -355,7 +339,7 @@ where
                 continue;
             }
 
-            let new_cost = costmodel(k as u32, sublength as u32) + costs[j] as f64;
+            let new_cost = costmodel(k, sublength) + costs[j] as f64;
             debug_assert!(new_cost >= 0.0);
             if new_cost < costs[j + k] as f64 {
                 debug_assert!(k <= ZOPFLI_MAX_MATCH);
@@ -374,12 +358,12 @@ where
 /// `length_array`. The `length_array` must contain the optimal length to reach that
 /// byte. The path will be filled with the lengths to use, so its data size will be
 /// the amount of lz77 symbols.
-fn trace_backwards(size: usize, length_array: &[u16]) -> Vec<u16> {
+fn trace(size: usize, length_array: &[u16]) -> Vec<u16> {
     let mut index = size;
     if size == 0 {
         return vec![];
     }
-    let mut path = vec![]; // TODO: with capacity
+    let mut path = Vec::with_capacity(index);
 
     while index > 0 {
         let lai = length_array[index];
@@ -391,8 +375,6 @@ fn trace_backwards(size: usize, length_array: &[u16]) -> Vec<u16> {
         index -= laiu;
     }
 
-    /* Mirror result. */
-    path.reverse();
     path
 }
 
@@ -408,8 +390,8 @@ fn trace_backwards(size: usize, length_array: &[u16]) -> Vec<u16> {
 /// returns the cost that was, according to the `costmodel`, needed to get to the end.
 ///     This is not the actual cost.
 #[allow(clippy::too_many_arguments)] // Not feasible to refactor in a more readable way
-fn lz77_optimal_run<F, C>(
-    s: &mut ZopfliBlockState<C>,
+fn lz77_optimal_run<F: Fn(usize, u16) -> f64, C: Cache>(
+    lmc: &mut C,
     in_data: &[u8],
     instart: usize,
     inend: usize,
@@ -417,14 +399,11 @@ fn lz77_optimal_run<F, C>(
     store: &mut Lz77Store,
     h: &mut ZopfliHash,
     costs: &mut Vec<f32>,
-) where
-    F: Fn(u32, u32) -> f64,
-    C: Cache,
-{
-    let (cost, length_array) = get_best_lengths(s, in_data, instart, inend, costmodel, h, costs);
-    let path = trace_backwards(inend - instart, &length_array);
-    store.follow_path(in_data, instart, inend, path, s);
-    debug_assert!(cost < f64::MAX);
+) {
+    let (cost, length_array) = get_best_lengths(lmc, in_data, instart, inend, costmodel, h, costs);
+    let path = trace(inend - instart, &length_array);
+    store.follow_path(in_data, instart, inend, path, lmc);
+    debug_assert!(cost < f64::INFINITY);
 }
 
 /// Does the same as `lz77_optimal`, but optimized for the fixed tree of the
@@ -435,21 +414,17 @@ fn lz77_optimal_run<F, C>(
 /// using with a fixed tree.
 /// If `instart` is larger than `0`, it uses values before `instart` as starting
 /// dictionary.
-pub fn lz77_optimal_fixed<C>(
-    s: &mut ZopfliBlockState<C>,
+pub fn lz77_optimal_fixed<C: Cache>(
+    lmc: &mut C,
     in_data: &[u8],
     instart: usize,
     inend: usize,
     store: &mut Lz77Store,
-) where
-    C: Cache,
-{
-    s.blockstart = instart;
-    s.blockend = inend;
+) {
     let mut h = ZopfliHash::new();
     let mut costs = Vec::with_capacity(inend - instart);
     lz77_optimal_run(
-        s,
+        lmc,
         in_data,
         instart,
         inend,
@@ -463,22 +438,20 @@ pub fn lz77_optimal_fixed<C>(
 /// Calculates lit/len and dist pairs for given data.
 /// If `instart` is larger than 0, it uses values before `instart` as starting
 /// dictionary.
-pub fn lz77_optimal<C>(
-    s: &mut ZopfliBlockState<C>,
+pub fn lz77_optimal<C: Cache>(
+    lmc: &mut C,
     in_data: &[u8],
     instart: usize,
     inend: usize,
-    numiterations: u8,
-) -> Lz77Store
-where
-    C: Cache,
-{
+    max_iterations: u64,
+    max_iterations_without_improvement: u64,
+) -> Lz77Store {
     /* Dist to get to here with smallest cost. */
     let mut currentstore = Lz77Store::new();
     let mut outputstore = currentstore.clone();
 
     /* Initial run. */
-    currentstore.greedy(s, in_data, instart, inend);
+    currentstore.greedy(lmc, in_data, instart, inend);
     let mut stats = SymbolStats::default();
     stats.get_statistics(&currentstore);
 
@@ -487,20 +460,22 @@ where
 
     let mut beststats = SymbolStats::default();
 
-    let mut bestcost = f64::MAX;
+    let mut bestcost = f64::INFINITY;
     let mut lastcost = 0.0;
     /* Try randomizing the costs a bit once the size stabilizes. */
     let mut ran_state = RanState::new();
-    let mut lastrandomstep = -1;
+    let mut lastrandomstep = u64::MAX;
 
     /* Do regular deflate, then loop multiple shortest path runs, each time using
     the statistics of the previous run. */
     /* Repeat statistics with each time the cost model from the previous stat
     run. */
-    for i in 0..numiterations as i32 {
+    let mut current_iteration: u64 = 0;
+    let mut iterations_without_improvement: u64 = 0;
+    loop {
         currentstore.reset();
         lz77_optimal_run(
-            s,
+            lmc,
             in_data,
             instart,
             inend,
@@ -512,30 +487,39 @@ where
         let cost = calculate_block_size(&currentstore, 0, currentstore.size(), BlockType::Dynamic);
 
         if cost < bestcost {
+            iterations_without_improvement = 0;
             /* Copy to the output store. */
             outputstore = currentstore.clone();
             beststats = stats;
             bestcost = cost;
 
-            debug!("Iteration {}: {} bit", i, cost);
+            debug!("Iteration {}: {} bit", current_iteration, cost);
         } else {
-            trace!("Iteration {}: {} bit", i, cost);
+            iterations_without_improvement += 1;
+            trace!("Iteration {}: {} bit", current_iteration, cost);
+            if iterations_without_improvement >= max_iterations_without_improvement {
+                break;
+            }
+        }
+        current_iteration += 1;
+        if current_iteration >= max_iterations {
+            break;
         }
         let laststats = stats;
         stats.clear_freqs();
         stats.get_statistics(&currentstore);
-        if lastrandomstep != -1 {
+        if lastrandomstep != u64::MAX {
             /* This makes it converge slower but better. Do it only once the
             randomness kicks in so that if the user does few iterations, it gives a
             better result sooner. */
             stats = add_weighed_stat_freqs(&stats, 1.0, &laststats, 0.5);
             stats.calculate_entropy();
         }
-        if i > 5 && (cost - lastcost).abs() < f64::EPSILON {
+        if current_iteration > 5 && (cost - lastcost).abs() < f64::EPSILON {
             stats = beststats;
             stats.randomize_stat_freqs(&mut ran_state);
             stats.calculate_entropy();
-            lastrandomstep = i;
+            lastrandomstep = current_iteration;
         }
         lastcost = cost;
     }
